// Copyright 2009, 2010, 2011 Christopher Smith
// 
// This file is part of smithers.algorithms.
// 
// smithers.algorithms is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.algorithms is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.algorithms. If not, see <http://www.gnu.org/licenses/>.

package smithers.algorithms;

/**
 * Implements a collection of useful vector related algorithms.
 */
public class Vector {
    
    private Vector() {}
    
    /**
     * Computes the sum of vectors.
     * 
     * @param vectors some vectors
     * @return the sum of the vectors
     */
    public static float[] sum(float[]... vectors) {
        final int m = vectors.length;
        final int n = vectors[0].length;
        float[] result = new float[n];
        for (int i=0; i<n; i++) {
            result[i] = 0;
            for (int j=0; j<m; j++)
                result[i] += vectors[j][i];
        }
        return result;
    }
    
    /**
     * Computes the difference between two vectors.
     * 
     * @param vector1 a vector
     * @param vector2 another vector
     * @return {@code vector1 - vector2}
     */
    public static float[] difference(float[] vector1, float[] vector2) {
        final int n = vector1.length;
        float[] result = new float[n];
        for (int i=0; i<n; i++)
            result[i] = vector1[i] - vector2[i];
        return result;
    }
    
    /**
     * Computes the scalar product of two vectors.
     * 
     * @param vector1 a vector
     * @param vector2 another vector
     * @return {@code vector1}&sdot;{@code vector2}
     */
    public static float scalarProduct(float[] vector1, float[] vector2) {
        final int n = vector1.length;
        float result = 0;
        for (int i=0; i<n; i++)
            result += vector1[i] * vector2[i];
        return result;
    }
    
    /**
     * Computes the vector product of two vectors. The supplied vectors must
     * both have length 3,
     * 
     * @param vector1 a vector
     * @param vector2 another vector
     * @return {@code vector1}&times;{@code vector2}
     */
    public static float[] vectorProduct(float[] vector1, float[] vector2) {
        return new float[] {vector1[1] * vector2[2] - vector1[2] * vector2[1],
                            vector1[2] * vector2[0] - vector1[0] * vector2[2],
                            vector1[0] * vector2[1] - vector1[1] * vector2[0]};
    }
    
    /**
     * Calculates the scalar triple product of 3 vectors.
     * 
     * @param p the first vector
     * @param q the second vector
     * @param r the third vector
     * @return {@code p}&sdot;{@code (p}&times;{@code q)}
     */
    public static float tripleProduct(float[] p, float[] q, float[] r) {
        return scalarProduct(p, vectorProduct(q, r));
    }
    
    /**
     * Normalises a vector.
     * 
     * @param vector a vector
     * @return a unit vector parallel to {@code vector}
     */
    public static float[] normalise(float[] vector) {
        float temp = (float)Math.sqrt(scalarProduct(vector, vector));
        final int n = vector.length;
        float[] result = new float[n];
        for (int i=0; i<n; i++)
            result[i] = vector[i] / temp;
        return result;
    }
    
}
