// Copyright 2011 Christopher Smith
// 
// This file is part of smithers.extras.
// 
// smithers.extras is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.extras is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.extras. If not, see <http://www.gnu.org/licenses/>.

package smithers.extras;

import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.io.IOException;
import javax.imageio.ImageIO;

/**
 * Class to perform operations on images.
 */
public class ImageTool {
    
    private static class StringBox {
        
        public String string;
        
        public String getNextCommand() {
            if (string == null)
                return null;
            String[] temp = string.split(" +", 2);
            string = temp.length > 1 ? temp[1] : null;
            return temp[0];
        }
        
    }
    
    private static final int BYTE_MASK = 0xFF;
    private static final int RED_SHIFT = 16;
    private static final int GREEN_SHIFT = 8;
    private static final int BLUE_SHIFT = 0;
    
    private BufferedImage image = null;
    
    /**
     * Private constructor.
     */
    private ImageTool() {}
    
    private void loadCommand(StringBox line) {
        if (line.string == null) {
            System.out.println("Expected: filename");
            return;
        }
        try {
            image = ImageIO.read(new File(line.string));
        } catch (IOException e) {
            System.out.println("Error loading file: " + e.getMessage());
        }
    }
    
    private void saveCommand(StringBox line) {
        String format = line.getNextCommand();
        if (format == null || format.equals("")) {
            System.out.println("Expected: format, filename");
            return;
        }
        if (line.string == null) {
            System.out.println("Expected: filename");
            return;
        }
        try {
            ImageIO.write(image, format, new File(line.string));
        } catch (IOException e) {
            System.out.println("Error saving file: " + e.getMessage());
        }
    }
    
    private void convolveCommand(StringBox line) {
        String kernalName = line.getNextCommand();
        if (kernalName == null || kernalName.equals("")) {
            System.out.println("Expected: kernal");
            return;
        } else if (kernalName.equals("gaussian")) {
            String sigmaString = line.getNextCommand();
            if (sigmaString == null) {
                System.out.println("Expected: sigma");
                return;
            }
            double sigma;
            try {
                sigma = Double.parseDouble(sigmaString);
            } catch (NumberFormatException e) {
                System.out.println("Expected: sigma");
                return;
            }
            
            int w = image.getWidth(), h = image.getHeight();
            int[] data = new int[w * h];
            double[] redInter = new double[w * h];
            double[] greenInter = new double[w * h];
            double[] blueInter = new double[w * h];
            image.getRGB(0, 0, w, h, data, 0, w);
            
            double sqrt2pir2 = Math.sqrt(2 * Math.PI * sigma * sigma);
            double[] factors = new double[w > h ? w : h];
            for (int k=0; k<factors.length; k++)
                factors[k] = Math.exp(-(k*k) / (2*sigma*sigma)) / sqrt2pir2;
            
            // Convolve horizontally
            for (int i=0; i<h; i++) {
                for (int j=0; j<w; j++) {
                    double r=0.0, g=0.0, b=0.0;
                    for (int k=0; k<w; k++) {
                        double factor = factors[k>j ? k-j : j-k];
                        r += ((data[i*w+k] >> RED_SHIFT) & BYTE_MASK) * factor;
                        g += ((data[i*w+k] >> GREEN_SHIFT) & BYTE_MASK) * factor;
                        b += ((data[i*w+k] >> BLUE_SHIFT) & BYTE_MASK) * factor;
                    }
                    redInter[i*w+j] = r;
                    greenInter[i*w+j] = g;
                    blueInter[i*w+j] = b;
                }
            }
            
            // Convolve vertically
            for (int i=0; i<h; i++) {
                for (int j=0; j<w; j++) {
                    double r=0.0, g=0.0, b=0.0;
                    for (int k=0; k<h; k++) {
                        double factor = factors[k>i ? k-i : i-k];
                        r += redInter[k*w+j] * factor;
                        g += greenInter[k*w+j] * factor;
                        b += blueInter[k*w+j] * factor;
                    }
                    data[i*w+j] = ((r >= 255 ? 255 : (int)(r+0.5)) << RED_SHIFT) |
                                    ((g >= 255 ? 255 : (int)(g+0.5)) << GREEN_SHIFT) |
                                    ((b >= 255 ? 255 : (int)(b+0.5)) << BLUE_SHIFT);
                }
            }
            
            image.setRGB(0, 0, w, h, data, 0, w);
        } else {
            System.out.println("Unknown kernal: " + kernalName);
        }
    }
    
    private void helpCommand(StringBox line) {
        String command = line.getNextCommand();
        
        if (command == null || command.equals("")) {
            System.out.println("Commands:");
            System.out.println("load, save, convolve, exit, help");
            
        } else if (command.equals("load")) {
            System.out.println("load filename");
            System.out.println("    Loads from the given file.");
            
        } else if (command.equals("save")) {
            System.out.println("save format filename");
            System.out.println("    Saves to the given file in the given format.");
            
        } else if (command.equals("convolve")) {
            System.out.println("convolve gaussian sigma");
            System.out.println("    Convolve the image with a gaussian filter of the given size.");
            
        } else if (command.equals("exit")) {
            System.out.println("exit");
            System.out.println("    Exits the program.");
            
        } else if (command.equals("help")) {
            command = line.getNextCommand();
            
            if (command == null || command.equals("")) {
                System.out.println("help");
                System.out.println("    Lists the available commands.");
                System.out.println("help command");
                System.out.println("    Gives help about the given command.");
            } else {
                System.out.println("help " + command);
                System.out.println("    Gives help about the " + command + " command.");
            }
            
        } else {
            System.out.println(command);
            System.out.println("    There is no such command as " + command + ".");
        }
    }
    
    /**
     * Main method, run with no arguments and type 'help' for commands.
     * 
     * @param args the arguments
     */
    public static void main(String[] args) throws IOException {
        if (args.length >= 1 && (args[0].equals("-h") || args[0].equals("--help"))) {
            System.out.println("smithers.extras.ImageTool -- performs operations on images.");
            System.out.println("Usage:");
            System.out.println("        java smithers.extras.ImageTool [-h|--help]");
            System.out.println("If -h or --help is given, print this help and exit. Otherwise interpret commands sequentially.");
            System.out.println("At the program's prompt, type `help' to list commands, and `help command' for help on a specific command");
            return;
        }
        
        ImageTool tool = new ImageTool();
        
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        StringBox line = new StringBox();
        boolean interactive = System.console() != null;
        
        if (interactive) {
            System.out.print("$ ");
            System.out.flush();
        }
        while ((line.string = in.readLine()) != null) {
            String command = line.getNextCommand();
            
            if (command.equals("load")) {
                tool.loadCommand(line);
            } else if (command.equals("save")) {
                tool.saveCommand(line);
            } else if (command.equals("convolve")) {
                tool.convolveCommand(line);
            } else if (command.equals("exit")) {
                break;
            } else if (command.equals("help")) {
                tool.helpCommand(line);
            } else {
                System.out.println("Unknown command");
            }
            if (interactive) {
                System.out.print("$ ");
                System.out.flush();
            }
        }
    }
    
}
