// Copyright 2009 Christopher Smith
// 
// This file is part of smithers.extras.
// 
// smithers.extras is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.extras is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.extras. If not, see <http://www.gnu.org/licenses/>.

package smithers.extras;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

/**
 * Implements Lindenmayer systems and displays output with turtle graphics.
 */
public class Lindenmayer {
    
    private String state;
    private Map<String, String> rules;
    
    /**
     * Creates a new Lindenmayer system.
     * 
     * @param axiom the initial state
     * @param rules the replacement rules
     */
    public Lindenmayer(String axiom, Map<String, String> rules) {
        this.state=axiom;
        this.rules=rules;
    }
    
    /**
     * Gets the current state.
     * 
     * @return the current state
     */
    public String getState() {
        return state;
    }
    
    /**
     * Performs one step of the evolution.
     */
    public void nextStep() {
        for (int i=0; i<state.length(); i++) {
            for (String key : rules.keySet()) {
                if (state.startsWith(key, i)) {
                    state = state.substring(0, i) + rules.get(key) + state.substring(i+key.length());
                    i+=rules.get(key).length()-1;
                    break;
                }
            }
        }
    }
    
    /**
     * Creates a Lindenmayer system, evolves it for a number of steps, and
     * outputs as a string.
     * 
     * @param axiom the initial state of the system
     * @param rules the rules for the system
     * @param steps the number of steps to run
     */
    public static String evolveToString(String axiom, Map<String, String> rules, int steps) {
        Lindenmayer l = new Lindenmayer(axiom, rules);
        for (int i=0; i<steps; i++) l.nextStep();
        return l.state;
    }
    
    /**
     * Creates a Lindenmayer system, evolves it for a number of steps, and
     * outputs using turtle graphics, saving as an SVG file.
     * 
     * @param axiom    the initial state of the system
     * @param rules    the rules for the system
     * @param steps    the number of steps to run
     * @param angle    the angle to turn by
     * @param distance the distance to move
     * @param x        the initial x-coordinate
     * @param y        the initial y-coordinate
     * @param dir      the initial direction
     * @param width    the width of the image
     * @param height   the height of the image
     * @param fileName the filename to save under
     */
    public static void evolveToTG(String axiom, Map<String, String> rules, int steps, double angle, double distance, double x, double y, double dir, double width, double height, String fileName) throws IOException {
        Lindenmayer l = new Lindenmayer(axiom, rules);
        for (int i=0; i<steps; i++) l.nextStep();
        new TurtleGraphics(angle, distance).doTurtle(l.state, x, y, dir, width, height, fileName);
    }
    
    /**
     * Tester. Set the first argument to "-s" for a sierpinksi triangle, or
     * "-p" for a penrose tiling.
     * 
     * @param args the arguments
     */
    public static void main(String[] args) throws IOException {
        String axiom;
        HashMap<String, String> rules = new HashMap<String, String>();
        
        if (args.length > 0) {
            if (args[0].equals("-s")) {
                axiom = "1F";
                rules.put("1F", "1F-2F+1F+2F-1F");
                rules.put("2F", "2F2F");
                evolveToTG(axiom, rules, 6, Math.PI*2/3, 5, 10, 283, 0, 340, 293, "sierpinski.svg");
                return;
            } else if (args[0].equals("-p")) {
                axiom = "[7]++[7]++[7]++[7]++[7]";
                rules.put("F", "");
                rules.put("6", "8F++9F----7F[-8F----6F]++");
                rules.put("7", "+8F--9F[---6F--7F]+");
                rules.put("8", "-6F++7F[+++8F++9F]-");
                rules.put("9", "--8F++++6F[+9F++++7F]--7F");
                evolveToTG(axiom, rules, 5, Math.PI/5, 20, 250, 250, Math.PI/2, 500, 500, "penrose.svg");
                return;
            }
        }
        axiom = "A";
        rules.put("A", "B");
        rules.put("B", "AB");
        Lindenmayer l = new Lindenmayer(axiom, rules);
        for (int i=0; i<5; i++) {
            l.nextStep();
            System.out.println(l.getState());
        }
    }
}
