// Copyright 2009 Christopher Smith
// 
// This file is part of smithers.extras.
// 
// smithers.extras is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.extras is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.extras. If not, see <http://www.gnu.org/licenses/>.

package smithers.extras;

import java.util.*;
import smithers.svg.*;

/**
 * Draws simple vector graphics using a 'turtle' and saves to an SVG file. The
 * turtle has 2 unchanging parameters, <emph>angle</emph> and
 * <emph>distance</emph> The turtle operates by reading a string of characters
 * and performing the following actions:
 * <table>
 * <hr> <td>Name</td> <td>Default character</td> <td>Action</td> </hr>
 * <tr> <td>Turn Left</td> <td>+</td> <td>Rotates the turtle left by
 * angle.</td> </tr>
 * <tr> <td>Turn Right</td> <td>-</td> <td>Rotates the turtle right by
 * angle.</td> </tr>
 * <tr> <td>Forwards</td> <td>F</td> <td>Moves the turtle forwards by distance,
 * leaving a line behind it.</td> </tr>
 * <tr> <td>Forwards Nodraw</td> <td>f</td> <td>Moves the turtle forwards by
 * distance.</td> </tr>
 * <tr> <td>Backwards</td> <td>B</td> <td>Moves the turtle backwards by
 * distance, leaving a line behind it.</td> </tr>
 * <tr> <td>Backwards Nodraw</td> <td>b</td> <td>Moves the turtle backwards by
 * distance.</td> </tr>
 * <tr> <td>Push</td> <td>[</td> <td>Pushes the turtle's current position and
 * direction onto the stack.</td> </tr>
 * <tr> <td>Pop</td> <td>]</td> <td>Retrieves the turtle's current position and
 * direction from the stack.</td> </tr>
 * </table>
 */
public class TurtleGraphics {
    
    private double angle;
    private double distance;
    private char turnLeft;
    private char turnRight;
    private char forwards;
    private char forwardsNodraw;
    private char backwards;
    private char backwardsNodraw;
    private char push;
    private char pop;
    
    /**
     * Creates a new turtle with the specified angle and distance and the
     * default command characters.
     * 
     * @param angle           the angle to turn by
     * @param distance        the distance to draw
     */
    public TurtleGraphics(double angle, double distance) {
        this(angle, distance, '-', '+', 'F', 'f', 'B', 'b', '[', ']');
    }
    
    /**
     * Creates a new turtle according to specified options.
     * 
     * @param angle           the angle to turn by (in radians)
     * @param distance        the distance to draw
     * @param turnLeft        the character to trigger turning left
     * @param turnRight       the character to trigger turning right
     * @param forwards        the character to trigger drawing forwards
     * @param forwardsNodraw  the character to trigger moving forwards
     * @param backwards       the character to trigger drawing backwards
     * @param backwardsNodraw the character to trigger moving backwards
     * @param push            the character to store the current position
     * @param pop             the character to retrieve the last stored
     *                        position
     */
    public TurtleGraphics(double angle, double distance, char turnLeft, char turnRight, char forwards, char forwardsNodraw, char backwards, char backwardsNodraw, char push, char pop) {
        this.angle=angle;
        this.distance=distance;
        this.turnLeft=turnLeft;
        this.turnRight=turnRight;
        this.forwards=forwards;
        this.forwardsNodraw=forwardsNodraw;
        this.backwards=backwards;
        this.backwardsNodraw=backwardsNodraw;
        this.push=push;
        this.pop=pop;
    }
    
    /**
     * Draws the turtle graphics and saves an svg file.
     * 
     * @param code     the code to draw from
     * @param x        the initial x-coordinate
     * @param y        the initial y-coordinate
     * @param dir      the initial direction
     * @param width    the width of the image
     * @param height   the height of the image
     * @param fileName the name of the file to save to
     */
    public void doTurtle(String code, double x, double y, double dir, double width, double height, String fileName) throws java.io.IOException {
        Deque<Double> xStack = new LinkedList<Double>(), yStack = new LinkedList<Double>(), dirStack = new LinkedList<Double>();
        char c;
        
        SVG s = new SVG(String.valueOf(width), String.valueOf(height));
        StringBuilder curCode = new StringBuilder().append("<path d=\" M ").append(x).append(" ").append(y);
        
        for (int i=0; i<code.length(); i++) {
            c = code.charAt(i);
            if (c==turnLeft) {
                dir += angle;
            } else if (c==turnRight) {
                dir -= angle;
            } else if (c==forwards) {
                x+=distance*Math.cos(dir);
                y-=distance*Math.sin(dir);
                curCode.append(" L ").append(x).append(" ").append(y);
            } else if (c==forwardsNodraw) {
                x+=distance*Math.cos(dir);
                y-=distance*Math.sin(dir);
                curCode.append(" M ").append(x).append(" ").append(y);
            } else if (c==backwards) {
                x-=distance*Math.cos(dir);
                y+=distance*Math.sin(dir);
                curCode.append(" L ").append(x).append(" ").append(y);
            } else if (c==backwardsNodraw) {
                x-=distance*Math.cos(dir);
                y+=distance*Math.sin(dir);
                curCode.append(" M ").append(x).append(" ").append(y);
            } else if (c==push) {
                xStack.push(new Double(x));
                yStack.push(new Double(y));
                dirStack.push(new Double(dir));
            } else if (c==pop) {
                curCode.append("\" stroke=\"black\" fill=\"none\" />");
                s.add(new SVGElementByCode(curCode.toString()));
                x = xStack.pop().doubleValue();
                y = yStack.pop().doubleValue();
                dir = dirStack.pop().doubleValue();
                curCode = new StringBuilder().append("<path d=\" M ").append(x).append(" ").append(y);
            }
        }
        curCode.append("\" stroke=\"black\" fill=\"none\" />");
        s.add(new SVGElementByCode(curCode.toString()));
        
        s.write(new java.io.File(fileName));
    }
    
}

