// Copyright 2009 Christopher Smith
// 
// This file is part of smithers.files.
// 
// smithers.files is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.files is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.files. If not, see <http://www.gnu.org/licenses/>.

package smithers.files;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

/**
 * Reads data from a CSV file.
 */
public class CSVReader {
    
    private BufferedReader in;
    private String currentRow = null;
    private int columns = 0;
    private String[] cells = null;
    
    /**
     * Creates a {@code CSVReader} which reads from the specified file.
     * 
     * @param fileName the name of the file to read from
     * @throws FileNotFoundException if the file is not found
     */
    public CSVReader(String fileName) throws FileNotFoundException {
        in = new BufferedReader(new FileReader(fileName));
    }
    
    /**
     * Creates a {@code CSVReader} which reads from the specified file.
     * 
     * @param file the file to read from
     * @throws FileNotFoundException if the file is not found
     */
    public CSVReader(File file) throws FileNotFoundException {
        in = new BufferedReader(new FileReader(file));
    }
    
    /**
     * Returns the current row of the table. This method returns the same value
     * as the most recent call to {@code nextRow()}, if there was no such call,
     * this returns {@code null}.
     * 
     * @return the current table row as a single {@code String}
     */
    public String currentRow() {
        return currentRow;
    }
    
    /**
     * Returns the number of columns in the current row of the table.
     * 
     * @return the number of columns in the current table row
     */
    public int numColumns() {
        return columns;
    }
    
    /**
     * Returns the specified cell of the current row of the table. This method
     * will accept an index from 0 to {@code numColumns()}.
     * 
     * @param i the column index
     * @return the value of the cell
     * @throws IndexOutOfBoundsException if {@code i<0 || i>= numColumns()}
     */
    public String getCell(int i) {
        if (i<0) throw new IndexOutOfBoundsException("Cell index must be >=0");
        if (i>=columns) throw new IndexOutOfBoundsException("Cell index must be <numColumns() (which is "+columns+").");
        return cells[i];
    }
    
    /**
     * Reads the next table row from the file. This method reads the next row
     * of data from the file, updates the {@code currentRow()}, {@code
     * numColumns()} and {@code getCell(int)} methods and returns the row as a
     * {@code String}. If the end of the file has already been reached, {@code
     * null} is returned. If the end of the file is reached before the end of
     * the row (i.e. when the quotes defining a cell are unclosed), the cell
     * (and row) are assumed to be complete.
     * 
     * @return the new table row as a single {@code String}
     * @throws IOException if one occurs
     */
    public String nextRow() throws IOException {
        String input;
        columns = 0;
        cells = null;
        
        currentRow = in.readLine();
        if (currentRow == null) {
            columns = 0;
            cells = null;
            return currentRow;
        }
        
        if ((quoteCount(currentRow) & 1) != 0) {
            do {
                input = in.readLine();
                if (input == null) break;
                currentRow += '\n' + input;
            } while ((quoteCount(input) & 1) == 0);
        }
        
        countColumns();
        splitCells();
        return currentRow;
    }
    
    private static int quoteCount(String str) {
        int result = 0;
        for (int i=0; i<str.length(); i++) if (str.charAt(i) == '"') result++;
        return result;
    }
    
    private void countColumns() {
        int columns = 1, quotes = 0;
        
        for (int i=0; i<currentRow.length(); i++) {
            switch (currentRow.charAt(i)) {
                case ',':
                    if (quotes == 0) columns++;
                    break;
                case '"':
                    quotes ^= 1;
                    break;
            }
        }
    }
    
    private void splitCells() {
        cells = new String[columns];
        int column = 0;
        StringBuilder temp = new StringBuilder();
        int quotes = 0;
        char c;
        
        for (int i=0; i<currentRow.length(); i++) {
            c = currentRow.charAt(i);
            switch (c) {
                case ',':
                    if (quotes == 0) {
                        cells[column] = temp.toString();
                        column++;
                        temp = new StringBuilder();
                    } else {
                        temp.append(',');
                    }
                    break;
                    
                case '"':
                    if (quotes == 0) {
                        quotes = 1;
                    } else if (currentRow.length() > i+1 && currentRow.charAt(i+1) == '"') {
                        temp.append('"');
                        i++;
                    } else {
                        quotes = 0;
                    }
                    break;
                    
                default:
                    temp.append(c);
            }
        }
        
        cells[column] = temp.toString();
    }
    
}
