// Copyright 2009, 2010, 2011 Christopher Smith
// 
// This file is part of smithers.java3d.
// 
// smithers.java3d is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.java3d is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.java3d. If not, see <http://www.gnu.org/licenses/>.

package smithers.java3d;

import smithers.algorithms.Vector;

/**
 * Represents a line in 3-dimensional space.
 */
public class Line3D {
    
    private float[] base;
    private float[] vec;
    
    /**
     * Constructs the x-axis.
     */
    public Line3D() {
        this(0, 0, 0, 1, 0, 0);
    }
    
    /**
     * Constructs a line through the origin in the given direction. This could
     * also be thought of as a line through the origin and the specified point.
     * 
     * @param dx the relative change in x
     * @param dy the relative change in y
     * @param dz the relative change in z
     */
    public Line3D(float dx, float dy, float dz) {
        this(0, 0, 0, dx, dy, dz);
    }
    
    /**
     * Constructs a line through the given point in the given direction.
     * 
     * @param x0 the x coordinate of the starting point
     * @param y0 the y coordinate of the starting point
     * @param z0 the z coordinate of the starting point
     * @param dx the relative change in x
     * @param dy the relative change in y
     * @param dz the relative change in z
     */
    public Line3D(float x0, float y0, float z0, float dx, float dy, float dz) {
        if (dx==0.0 && dy==0.0 && dz==0.0) throw new IllegalArgumentException("The line is degenerate.");
        base = new float[] {x0, y0, z0};
        vec = new float[] {dx, dy, dz};
    }
    
    /**
     * Constructs a line through the given point in the given direction.
     * 
     * @param base the starting point
     * @param vec the direction vector
     */
    public Line3D(float[] base, float[] vec) {
        if (base.length != 3 || vec.length != 3) throw new IllegalArgumentException("base and vec must each have 3 components.");
        if (vec[0] == 0 && vec[1] == 0 && vec[2] == 0) throw new IllegalArgumentException("The line is degenerate.");
        this.base = base;
        this.vec = vec;
    }
    
    /**
     * Get a point on the line.
     * 
     * @return a point on the line
     */
    public float[] getBase() {
        return base;
    }
    
    /**
     * Get the direction in which the line points.
     * 
     * @return a vector parallel to this line
     */
    public float[] getVector() {
        return vec;
    }
    
    /**
     * Get another point on this line. The point returned is a point on the
     * line disinct from the one returned by {@link #getBase()}.
     * 
     * @return a point on this line, distinct from {@link #getBase()}
     */
    public float[] getHead() {
        return new float[] {base[0]+vec[0], base[1]+vec[1], base[2]+vec[2]};
    }
    
    /**
     * Calculates the shortest distance from this line to the given point.
     * 
     * @param p the point
     * @return the distance
     */
    public float distance(float[] p) {
        if (p.length != 3) throw new IllegalArgumentException("p must have length 3.");
        float[] temp = Vector.vectorProduct(Vector.normalise(vec), Vector.difference(p, base));
        return Vector.scalarProduct(temp, temp);
    }
    
    /**
     * Calculates the point of intersection of this line with a plane. The plane
     * is taken through the given point and normal to the given vector.
     * 
     * @param p a point on the plane
     * @param n the normal to the plane
     * @return the point of intersection
     */
    public float[] intersectPlane(float[] p, float[] n) {
        if (p.length != 3 || n.length != 3) throw new IllegalArgumentException("p and n must have length 3.");
        float temp = Vector.scalarProduct(vec, n);
        if (temp == 0) throw new IllegalArgumentException("This line is parallel to the plane");
        temp = Vector.scalarProduct(Vector.difference(p, base), n) / temp;
        float[] result = new float[3];
        for (int i=0; i<3; i++)
            result[i] = base[i] + vec[i] * temp;
        return result;
    }
    
}
