// Copyright 2009, 2010, 2011 Christopher Smith
// 
// This file is part of smithers.java3d.
// 
// smithers.java3d is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.java3d is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.java3d. If not, see <http://www.gnu.org/licenses/>.

package smithers.java3d;

import java.util.Arrays;

/**
 * Represents a polygon in 3D space. This class represents a polygon by a
 * sequence of vertices. No guarantee is made that these are coplanar (in fact,
 * such a guarantee would be impossible in the presence of floating-point
 * rounding errors), but most uses of this class will only make sense for
 * (almost) planar polgons.
 */
public class Polygon3D {
    
    private float[][] points;
    private int n;
    
    /**
     * Constructs a new, empty {@code Polygon3D}.
     */
    public Polygon3D() {
        this(3);
    }
    
    /**
     * Constructs a new, empty {@code Polygon3D}.
     * 
     * @param size the number of points the polygon should have space for
     */
    public Polygon3D(int size) {
        this.points = new float[size][];
        n = 0;
    }
    
    /**
     * Constructs a new {@code Polygon3D} from the specified points.
     * 
     * @param points the vertices of the new polygon
     */
    public Polygon3D(float[]... points) {
        n = points.length;
        this.points = Arrays.copyOf(points, n);
    }
    
    /**
     * Adds a new point to the end of the array of vertices.
     * 
     * @param point the new point
     */
    public void add(float[] point) {
        if (points.length <= n)
            points = Arrays.copyOf(points, 2*n);
        points[n] = point;
        n++;
    }
    
    /**
     * Returns the array of points which make up this polygon.
     * 
     * @return the array of vertices of the polygon
     */
    public float[][] points() {
        return Arrays.copyOf(points, n);
    }
    
    /**
     * Computes the "centre" of this polygon. The point returned is the centre
     * of mass of the vertices of the polygon.
     * 
     * @return the centroid of the polygon's vertices
     */
    public float[] vertexCentroid() {
        float[] result = {0, 0, 0};
        for (int i=0; i<n; i++) {
            result[0] += points[i][0];
            result[1] += points[i][1];
            result[2] += points[i][2];
        }
        result[0] /= n;
        result[1] /= n;
        result[2] /= n;
        return result;
    }
    
    /**
     * Scales the polygon by the given factor.
     * 
     * @param factor the scale factor to use
     * @return the scaled polygon
     */
    public Polygon3D scale(float factor) {
        float[] c = vertexCentroid();
        float[][] result = new float[n][3];
        for (int i=0; i<n; i++) for (int j=0; j<3; j++)
            result[i][j] = (1-factor) * c[j] + factor * points[i][j];
        return new Polygon3D(result);
    }
    
}
