// Copyright 2009, 2010, 2011 Christopher Smith
// 
// This file is part of smithers.java3d.
// 
// smithers.java3d is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.java3d is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.java3d. If not, see <http://www.gnu.org/licenses/>.

package smithers.java3d;

import smithers.algorithms.Matrix;

/**
 * Represents a transformation between two real vector spaces. For example,
 * from 3-dimensional space to 2-dimensional space. This class is general
 * enough to include all affine linear transformations, as well as perspective
 * transformations.
 */
public class Transformation {
    
    private final int n, m;
    private float[][] data;
    private float[][] l = null, q = null;
    
    private Transformation(float[][] data) {
        this.n = data.length - 1;
        this.m = data[0].length - 1;
        this.data = data;
    }
    
    /**
     * Constructs a new matrix representing an identity/projection/inclusion.
     * If {@code n==m} this is an identity transformation, if {@code n<m} this
     * is a projection onto the first {@code n} coordinates and if {@code n>m}
     * then this is an inclusion of the subspace specified by the first {@code
     * m} coordinates.
     * 
     * @param n the dimension of the space to map to
     * @param m the dimension of the space to map from
     */
    public Transformation(int n, int m) {
        if (n < 0 || m < 0) throw new IllegalArgumentException("Dimensions must be non-negative.");
        this.n = n;
        this.m = m;
        this.data = new float[n+1][m+1];
        int min = n < m ? n : m;
        for (int i=0; i<min; i++)
            data[i][i] = 1;
        data[n][m] = 1;
    }
    
    /**
     * Creates a perspective transformation. The returned transformation maps
     * from {@code m} dimensions down to {@code m-1}. The transformation is to
     * project onto the first {@code m-1} coordinates and then divide by the
     * final coordinate.
     * 
     * @param m the dimension to map from
     * @return the perspective transformation
     */
    public static Transformation perspective(int m) {
        Transformation result = new Transformation(m-1, m);
        result.data[m-1][m-1] = 1;
        result.data[m-1][m] = 0;
        return result;
    }
    
    public static Transformation translation(float... vector) {
        int n = vector.length;
        Transformation result = new Transformation(n, n);
        for (int i=0; i<n; i++)
            result.data[i][n] = vector[i];
        return result;
    }
    
    public static Transformation scale(int n, float scale) {
        Transformation result = new Transformation(n, n);
        result.data[n][n] = 1/scale;
        return result;
    }
    
    /**
     * Creates a rotation transformation. The returned transformation is a
     * rotation of {@code n}-dimensional space which rotates the axis indexed
     * by {@code axis1} towards the axis indexed by {@code axis2} by an angle
     * of {@code theta} radians.
     * 
     * @param n     the number of dimensions
     * @param axis1 the axis to rotate from
     * @param axis2 the axis to rotate towards
     * @param theta the angle to rotate by
     * @return the rotation
     */
    public static Transformation rotation(int n, int axis1, int axis2, float theta) {
        if (axis1 >= n || axis2 >= n || axis1 == axis2) throw new IllegalArgumentException("Invalid axes.");
        Transformation result = new Transformation(n, n);
        result.data[axis1][axis1] = result.data[axis2][axis2] = (float)Math.cos(theta);
        result.data[axis1][axis2] = -(result.data[axis2][axis1] = (float)Math.sin(theta));
        return result;
    }
    
    public static Transformation rotation2D(float theta) {
        return new Transformation(new float[][] {{(float)Math.cos(theta), -(float)Math.sin(theta), 0},
                                                 {(float)Math.sin(theta), (float)Math.cos(theta), 0},
                                                 {0, 0, 1}});
    }
    
    public static Transformation rotation3DFromIntrinsicZXZ(float theta1, float theta2, float theta3) {
        float c1 = (float)Math.cos(theta1);
        float s1 = (float)Math.sin(theta1);
        float c2 = (float)Math.cos(theta2);
        float s2 = (float)Math.sin(theta2);
        float c3 = (float)Math.cos(theta3);
        float s3 = (float)Math.sin(theta3);
        
        return new Transformation(new float[][] {{-s1*c2*s3 + c1*c3, -s1*c2*c3 - c1*s3, s1*s2, 0},
                                                 {c1*c2*s3 + s1*c3, c1*c2*c3 - s1*s3, -c1*s2, 0},
                                                 {s2*s3, s2*c3, c2, 0},
                                                 {0, 0, 0, 1}});
    }
    
    public static Transformation rotation3DFromIntrinsicXYZ(float theta1, float theta2, float theta3) {
        float c1 = (float)Math.cos(theta1);
        float s1 = (float)Math.sin(theta1);
        float c2 = (float)Math.cos(theta2);
        float s2 = (float)Math.sin(theta2);
        float c3 = (float)Math.cos(theta3);
        float s3 = (float)Math.sin(theta3);
        
        return new Transformation(new float[][] {{c2*c3, -c2*s3, s2, 0},
                                                 {s1*s2*c3 + c1*s3, -s1*s2*s3 + c1*c3, -s1*c2, 0},
                                                 {-c1*s2*c3 + s1*s3, c1*s2*s3 + s1*c3, c1*c2, 0},
                                                 {0, 0, 0, 1}});
    }
    
    /**
     * Computes the transformation obtained by performing the transformation
     * {@code mat1} after the transformation {@code mat2}. This is in keeping
     * with the order in the mathematical meaning of the composition {@code
     * mat1} &circ; {@code mat2}. {@code mat1} must map from the same number of
     * dimensiona that {@code mat2} maps to.
     * 
     * @param mat1 the operation to apply second
     * @param mat2 the operation to apply first
     * @return the composite transformation
     */
    public static Transformation composition(Transformation mat1, Transformation mat2) {
        if (mat1.m != mat2.n) throw new IllegalArgumentException("Dimensions do not match.");
        int n = mat1.n, m = mat2.m, l = mat1.m;
        Transformation result = new Transformation(n, m);
        for (int i=0; i<=n; i++) for (int j=0; j<=m; j++) {
            result.data[i][j] = 0;
            for (int k=0; k<=l; k++)
                result.data[i][j] += mat1.data[i][k] * mat2.data[k][j];
        }
        return result;
    }
    
    /**
     * Prepends the given operation onto this one. The given operation must be
     * square of the same dimension that this one maps from.
     * 
     * @param other the transformation to be prepended
     */
    public void preCompose(Transformation other) {
        if (m != other.n || m != other.m) throw new IllegalArgumentException("Dimensions do not match.");
        float[] copy = new float[m+1];
        for (int i=0; i<=n; i++) {
            for (int j=0; j<=m; j++)
                copy[j] = data[i][j];
            for (int j=0; j<=m; j++) {
                data[i][j] = 0;
                for (int k=0; k<=m; k++)
                    data[i][j] += copy[k] * other.data[k][j];
            }
        }
        l = q = null;
    }
    
    /**
     * Appends the given operation onto this one. The given operation must be
     * square of the same dimension that this one maps to.
     * 
     * @param other the transformation to be appended
     */
    public void postCompose(Transformation other) {
        if (n != other.n || n != other.m) throw new IllegalArgumentException("Dimensions do not match.");
        float[] copy = new float[n+1];
        for (int j=0; j<=m; j++) {
            for (int i=0; i<=n; i++)
                copy[i] = data[i][j];
            for (int i=0; i<=n; i++) {
                data[i][j] = 0;
                for (int k=0; k<=n; k++)
                    data[i][j] += other.data[i][k] * copy[k];
            }
        }
        l = q = null;
    }
    
    public float getFactor(float[] vector) {
        if (vector.length != m) throw new IllegalArgumentException("vector must have length "+m+".");
        
        float result = data[n][m];
        for (int j=0; j<m; j++)
            result += data[n][j] * vector[j];
        
        return result;
    }
    
    /**
     * Apply this transformation to a vector. The length of the vector must be
     * the same as the dimension this transformation maps from.
     * 
     * @param vector the vector to transform
     * @return the transformed vector
     */
    public float[] apply(float[] vector) {
        if (vector.length != m) throw new IllegalArgumentException("vector must have length "+m+".");
        
        float factor = data[n][m];
        for (int j=0; j<m; j++)
            factor += data[n][j] * vector[j];
        
        float[] result = new float[n];
        for (int i=0; i<n; i++) {
            result[i] = data[i][m];
            for (int j=0; j<m; j++)
                result[i] += data[i][j] * vector[j];
            result[i] /= factor;
        }
        
        return result;
    }
    
    /**
     * Apply this transformation to an array of vectors. Each element of {@code
     * vector} must be the same as the dimension this transformation maps from.
     * If {@code result} is not {@code null}, the result vectors will be stored
     * in it. In this case it must have the same length as {@code vector}, and
     * each element must have the length of the dimension that this
     * transformation maps to.
     * 
     * @param vector the vectors to transform
     * @param result if not {@code null}, the result will be stored here
     * @return the transformed vectors
     */
    public float[][] apply(float[][] vector, float result[][]) {
        int l = vector.length;
        if (result == null) result = new float[l][n];
        else if (result.length != l) throw new IllegalArgumentException("vector and result must have same length.");
        
        for (int k=0; k<l; k++) {
            if (vector[k].length != m) throw new IllegalArgumentException("Each element of vector must have length "+m+".");
            if (result[k].length != n) throw new IllegalArgumentException("Each element of result must have length "+n+".");
            
            float factor = data[n][m];
            for (int j=0; j<m; j++)
                factor += data[n][j] * vector[k][j];
            
            for (int i=0; i<n; i++) {
                result[k][i] = data[i][m];
                for (int j=0; j<m; j++)
                    result[k][i] += data[i][j] * vector[k][j];
                result[k][i] /= factor;
            }
        }
        
        return result;
    }
    
    public float[][] invert(float[] vector) {
        if (m < n) throw new RuntimeException("Transformation could not be inverted.");
        if (vector.length != n) throw new RuntimeException("vector must have length "+n+".");
        if (l == null) {
            l = new float[n+1][m+1];
            q = new float[m+1][m+1];
            Matrix.decomposeLQ(data, l, q);
        }
        
        int i, j, k;
        float[] intermediate = new float[m+1];
        for (i=0; i<=n; i++) {
            if (l[i][i] == 0) throw new RuntimeException("Transformation could not be inverted.");
            intermediate[i] = i<n ? vector[i] : 1;
            for (j=0; j<i; j++)
                intermediate[i] -= l[i][j] * intermediate[j];
            intermediate[i] /= l[i][i];
        }
        for (i=n+1; i<=m; i++)
            intermediate[i] = -10;
        
        float[][] result = new float[m-n+1][m];
        for (k=0; k<=m-n; k++) {
            if (k != 0)
                intermediate[n+k] = 10;
            float factor = 0;
            for (j=0; j<=m; j++)
                factor += intermediate[j] * q[j][m];
            for (i=0; i<m; i++) {
                result[k][i] = 0;
                for (j=0; j<=m; j++)
                    result[k][i] += intermediate[j] * q[j][i];
                result[k][i] /= factor;
            }
        }
        
        return result;
    }
    
}
