// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

/**
 * Class to aid implementations of {@code PuzzleSolver}.
 */
public abstract class AbstractPuzzleSolver<P extends Puzzle, Q extends PuzzleState<P>> implements PuzzleSolver<P, Q> {
    
    protected P puzzle;
    protected Q state;
    
    /**
     * Creates a new solver for the specified puzzle.
     * 
     * @param puzzle the puzzle to solve
     */
    public AbstractPuzzleSolver(P puzzle) {
        this.puzzle = puzzle.isMutable() ? (P) puzzle.makeImmutableCopy() : puzzle;
        state = makeState(puzzle);
    }
    
    /**
     * Creates a new solver to work from the specified state.
     * 
     * @param state the puzzle state to work from
     */
    public AbstractPuzzleSolver(Q state) {
        this.state = state;
        puzzle = state.getPuzzle();
    }
    
    /**
     * {@inheritDoc}
     */
    public P getPuzzle() {
        return puzzle;
    }
    
    /**
     * {@inheritDoc}
     */
    public Q getState() {
        return state;
    }
    
    /**
     * Makes a new state for the specified puzzle. Usually this will simply
     * call a constructor which takes {@code puzzle} as an argument.
     * 
     * @param puzzle the puzzle to create a state for
     * @return a state for the given puzzle
     */
    public abstract Q makeState(P puzzle);
    
    /**
     * Attempts to solve the current step of the puzzle. This method must
     * return {@code 0} if it made no changes, {@code 1} if it made a change to
     * its internal data, but no change to the state of the puzzle, or any
     * number greater than {@code 1} if a change has been made to the state.
     * 
     * @return the return value described above
     */
    public abstract int solveStep();
    
    /**
     * Advances to the next step of solving the puzzle.
     * 
     * @return {@code true} iff the solver has started a new 'pass'
     */
    public abstract boolean nextStep();
    
    /**
     * {@inheritDoc}
     */
    public boolean step() {
        boolean changed = true;
        int value;
        while (true) {
            value = solveStep();
            if (value > 1) {
                nextStep();
                return true;
            } else if (value == 1) {
                changed = true;
            }
            if (nextStep()) {
                if (!changed) return false;
                changed = false;
            }
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public boolean pass() {
        boolean changed = false;
        do changed |= solveStep() > 1;
        while (!nextStep());
        return changed;
    }
    
    /**
     * {@inheritDoc}
     */
    public boolean solve() {
        boolean changed = false;
        boolean passChange = true;
        int temp;
        while (true) {
            temp = solveStep();
            changed |= temp > 1;
            passChange |= temp >= 1;
            if (nextStep()) {
                if (!passChange) break;
                passChange = false;
            }
        }
        return changed;
    }
    
}
