// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

/**
 * Interface representing a solver for a certain type of logic puzzle. The
 * solver may assume that if any other object modifies the state then the
 * method {@code start()} will be called before the solving methods are used
 * again.
 */
public interface PuzzleSolver<P extends Puzzle, Q extends PuzzleState<P>> {
    
    /**
     * Returns the puzzle which is being solved.
     * 
     * @return the {@code Puzzle} which this {@code PuzzleSolver} relates to.
     */
    public P getPuzzle();
    
    /**
     * Returns the state of the solution attempt. This method should always
     * return the same object, with updates being made to it as the solution
     * progresses.
     * 
     * @return the {@code Puzzle} which this {@code State} relates to.
     */
    public Q getState();
    
    /**
     * Restarts the solving process. This method should set the solver's
     * internal status to reflect any changes made to the state.
     */
    public void start();
    
    /**
     * Performs a single step in the solving process. This method should do the
     * smallest amount of work which results in a change being made to
     * {@code puzzle}.
     * 
     * @return {@code true} iff a change was made
     */
    public boolean step();
    
    /**
     * Performs a single pass through the puzzle. This method should perform
     * some small amount of the solution to the puzzle. Typically, this method
     * should perform roughly a single pass through the clues of the puzzle, or
     * through one subcollection of the clues (such as the row or column
     * clues). The work done by this method should ideally lie somewhere
     * between {@code step} and {@code solve}. If no intermediate is available,
     * this should perform the same function as {@code step}.
     * 
     * Note that unlike the other solving methods, this method may return
     * {@code false} even when this solver can still make further deductions.
     * 
     * @return {@code true} iff a change was made
     */
    public boolean pass();
    
    /**
     * Attempts to completely solve the puzzle. If the puzzle cannot be solved,
     * this ahould solve as much as is possible. Calling this method should be
     * equivalent to {@code while (this.step());} and to
     * {@code while (this.pass());}
     * 
     * @return {@code true} iff a change was made
     */
    public boolean solve();
    
}
