// Copyright 2009, 2010 Christopher Smith
// 
// This file is part of smithers.twisty.
// 
// smithers.twisty is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.twisty is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.twisty. If not, see <http://www.gnu.org/licenses/>.

package smithers.twisty;

import java.util.Arrays;
import java.util.Random;
import smithers.java3d.Polygon3D;
import smithers.java3d.Transformation;

/*
Stickers
========
   +--   +--------+
   |  ---|25  6 22|
   |     | 0 --  4|
   |     |16  2 19|--+
+--------+--------+--------+--------+
|27  1 17|18  3 20|21  5 23|24  7 26|
|45  40  |   41   |42  43  |   44   |
|38  9 30|29 11 33|32 13 36|35 15 38|
+--------+--------+--------+--------+
   |     |28 10 31|--+
   |     | 8 -- 12|
   |  ---|37 14 34|
   +--   +--------+
 0-15: Edge bits
       ===0(2) -> top/bottom
       ===1(2) -> side
16-39: Corner bits
       ===1(3) -> top/bottom
       ===2(3) -> side TR/BL
       ===0(3) -> side TL/BR
40-45: Middle layer
       ===1(3) -> long bit
       ===2(3) -> full width
       ===0(3) -> short bit

Top locations
=============
 /--------\
/ 10  9  8 \
+11_      7|
| 0 ----_ 6|
| 1       5+
\  2  3  4 /
 \--------/

Bottom locations
================
 /--------\
/ 14 15 16 \
|13     _17+
|12_---- 18|
+23      19|
\ 22 21 20 /
 \--------/

Pieces
======
         +--------+
         |11  3 10|
         | 0     2|
         | 8  1  9|
+--------+--------+--------+--------+
|        |        |        |        |
|        |   16   |        |   17   |
|        |        |        |        |
+--------+--------+--------+--------+
         |12  5 13|
         | 4     6|
         |15  7 14|
         +--------+
*/

/**
 * Represents the twisty puzzle <a
 * href="http://en.wikipedia.org/wiki/Square_One_%28puzzle%29"><emph>Square
 * One</emph>/<emph>Cube 21</emph></a>.
 */
public class SquareOne implements AnimatedTwistyPuzzle {
    
    private static class SquareOneTwist extends Twist {
    }
    
    private static class TurningTwist extends SquareOneTwist {
        
        int top, bot;
        
        TurningTwist(int top, int bot) {
            this.top = (top%12+18)%12-6;
            this.bot = (bot%12+18)%12-6;
        }
        
    }
    
    private static class FlippingTwist extends SquareOneTwist {
        
        int which;
        
        FlippingTwist(int which) {
            this.which = which;
        }
        
    }
    
    private final int[] piece = new int[24];
    private final int[] location = new int[16];
    private final boolean[] midFlip = new boolean[2];
    private final int[] colours = new int[6];
    
    private boolean animating = false;
    private double aniTime=0, aniCurrent=0;
    private SquareOneTwist aniTwist=null;
    
    /**
     * Constructs a new Square One puzzle.
     */
    public SquareOne() {
        for (int i=0; i<8; i++) {
            piece[3*i] = i;
            piece[3*i+1] = i+8;
            piece[3*i+2] = i+8;
            location[i] = 3*i;
            location[i+8] = 3*i+1;
        }
        midFlip[0] = false;
        midFlip[1] = false;
        colours[0] = 0xffff9900; // Orange
        colours[1] = 0xffffffff; // White
        colours[2] = 0xff00ff00; // Green
        colours[3] = 0xff0000ff; // Blue
        colours[4] = 0xffffff00; // Yellow
        colours[5] = 0xffff0000; // Red
    }
    
    /**
     * {@inheritDoc}
     */
    public int stickerCount() {
        return 46;
    }
    
    /**
     * {@inheritDoc}
     */
    public int getColor(int index) {
        if (index < 0) {
            throw new IndexOutOfBoundsException("index must be >=0");
        } else if (index < 16) { //Edge piece
            if (index%2 == 0) {
                return colours[index<8 ? 0 : 5];
            } else {
                return colours[index/2%4+1];
            }
        } else if (index < 40) { //Corner piece
            if (index%3 == 1) {
                return colours[index<28 ? 0 : 5];
            } else if (index%3 == 2) {
                if (index<29) return colours[index/3-4];
                else return colours[index/3%4+1];
            } else {
                if (index<30) return colours[(index/3-1)%4+1];
                else return colours[index/3-9];
            }
        } else if (index < 46) { //Middle layer
            return colours[(index*2/3-2)%4+1];
        } else {
            throw new IndexOutOfBoundsException("index must be <46");
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public double getScale() {
        return 4;
    }
    
    /**
     * {@inheritDoc}
     */
    public Polygon3D getPolygon(int index) {
        double r1 = 2/Math.sqrt(2+Math.sqrt(3));         // sec(PI/12)
        double r2 = Math.sqrt(2);                        // sec(PI/4)
        float y1 = (float)(2-Math.sqrt(3));              // tan(PI/12)
        float y2 = (float)(2/Math.sqrt(2+Math.sqrt(3))); // sec(PI/12)
        
        float[][] pts;
        int sideOfFlip=0; // 0 for front half, 1 for back half. Used if animating a flipping twist.
        double flippedness=0; // Used if animating a flipping twist or drawing a flipped middle layer.
        
        if (index < 0) {
            throw new IndexOutOfBoundsException("Index must be >=0");
        } else if (index < 40) {
            int loc = location[piece(index)];
            if(loc >= 12) {
                y2 = -y2;
                y1 = -y1;
            }
            
            if (loc >= 6 &&  loc < 12 || loc >= 18) sideOfFlip = 1;
            
            double aOffset = 0;
            if (animating && aniTwist instanceof TurningTwist) {
                TurningTwist tt = (TurningTwist)aniTwist;
                aOffset = (loc<12 ? -tt.top : tt.bot) * aniCurrent / aniTime * Math.PI / 6;
            }
            
            if (index < 16) { //Edge piece
                double a1, a2;
                float x1, x2, z1, z2;
                if (loc < 12) {
                    a1 = (loc-0.5)*Math.PI/6 + aOffset;
                    a2 = (loc+0.5)*Math.PI/6 + aOffset;
                } else {
                    a1 = (loc+0.5)*Math.PI/6 + aOffset;
                    a2 = (loc-0.5)*Math.PI/6 + aOffset;
                }
                x1 = (float)(-r1 * Math.cos(a1));
                z1 = (float)(r1 * Math.sin(a1));
                x2 = (float)(-r1 * Math.cos(a2));
                z2 = (float)(r1 * Math.sin(a2));
                if (index%2 == 0) { //Top
                    float[][] temp = {{0, y2, 0},
                                      {x2, y2, z2},
                                      {x1, y2, z1}};
                    pts = temp;
                } else { //Side
                    float[][] temp = {{x1, y2, z1},
                                      {x2, y2, z2},
                                      {x2, y1, z2},
                                      {x1, y1, z1}};
                    pts = temp;
                }
            } else { //Corner piece
                double a1, a2, a3;
                float x1, x2, x3, z1, z2, z3;
                if (loc < 12) {
                    a1 = (loc-0.5)*Math.PI/6 + aOffset;
                    a2 = (loc+0.5)*Math.PI/6 + aOffset;
                    a3 = (loc+1.5)*Math.PI/6 + aOffset;
                } else {
                    a1 = (loc+1.5)*Math.PI/6 + aOffset;
                    a2 = (loc+0.5)*Math.PI/6 + aOffset;
                    a3 = (loc-0.5)*Math.PI/6 + aOffset;
                }
                x1 = (float)(-r1 * Math.cos(a1));
                z1 = (float)(r1 * Math.sin(a1));
                x2 = (float)(-r2 * Math.cos(a2));
                z2 = (float)(r2 * Math.sin(a2));
                x3 = (float)(-r1 * Math.cos(a3));
                z3 = (float)(r1 * Math.sin(a3));
                if (index%3 == 1) { //Top
                    float[][] temp = {{0, y2, 0},
                                      {x3, y2, z3},
                                      {x2, y2, z2},
                                      {x1, y2, z1}};
                    pts = temp;
                } else if (index%3 == 2) { //Side TR/BL
                    float[][] temp = {{x1, y2, z1},
                                      {x2, y2, z2},
                                      {x2, y1, z2},
                                      {x1, y1, z1}};
                    pts = temp;
                } else { //Side TL/BR
                    float[][] temp = {{x2, y2, z2},
                                      {x3, y2, z3},
                                      {x3, y1, z3},
                                      {x2, y1, z2}};
                    pts = temp;
                }
            }
        } else if (index < 46) { //Middle layer
            double a1, a2, a3, a4;
            float x1, x2, x3, x4, z1, z2, z3, z4;
            int p = index<43 ? 0 : 1;
            sideOfFlip = p;
            if (midFlip[p]) flippedness += 1;
            
            a1 = (p - 1/12.) * Math.PI;
            a2 = (p + 3/12.) * Math.PI;
            a3 = (p + 9/12.) * Math.PI;
            a4 = (p + 11/12.) * Math.PI;
            x1 = (float)(-r1 * Math.cos(a1));
            z1 = (float)(r1 * Math.sin(a1));
            x2 = (float)(-r2 * Math.cos(a2));
            z2 = (float)(r2 * Math.sin(a2));
            x3 = (float)(-r2 * Math.cos(a3));
            z3 = (float)(r2 * Math.sin(a3));
            x4 = (float)(-r1 * Math.cos(a4));
            z4 = (float)(r1 * Math.sin(a4));
            
            if (index%3 == 1) { //Long bit
                float[][] temp = {{x1, y1, z1},
                                  {x2, y1, z2},
                                  {x2, -y1, z2},
                                  {x1, -y1, z1}};
                pts = temp;
            } else if (index%3 == 2) { //Full side
                float[][] temp = {{x2, y1, z2},
                                  {x3, y1, z3},
                                  {x3, -y1, z3},
                                  {x2, -y1, z2}};
                pts = temp;
            } else { //Short bit
                float[][] temp = {{x3, y1, z3},
                                  {x4, y1, z4},
                                  {x4, -y1, z4},
                                  {x3, -y1, z3}};
                pts = temp;
            }
        } else {
            throw new IndexOutOfBoundsException("index must be <46");
        }
        
        if (animating && aniTwist instanceof FlippingTwist && ((FlippingTwist)aniTwist).which == sideOfFlip)
            flippedness += aniCurrent / aniTime;
        
        if (flippedness != 0.0) {
            double angle = flippedness*Math.PI;
            
            Transformation trans = Transformation.rotation(3, 2, 0, (float)Math.PI/12);
            trans.postCompose(Transformation.rotation(3, 0, 1, (float)angle));
            trans.postCompose(Transformation.rotation(3, 0, 2, (float)Math.PI/12));
            for (int i=0; i<pts.length; i++)
                pts[i] = trans.apply(pts[i]);
        }
        
        return new Polygon3D(pts);
    }
    
    public Twist getTurn(int top, int bottom) {
        return new TurningTwist(top, bottom);
    }
    
    public Twist getFlip(int which) {
        return new FlippingTwist(which & 1);
    }
    
    /**
     * {@inheritDoc}
     */
    public Twist getTwist(int index, int dir) {
        if (index < 0) {
            throw new IndexOutOfBoundsException("Index must be >=0");
        } else if (index < 40) {
            int top=0, bot=0;
            dir = dir==0 ? 1 : dir==1 ? -1 : 6;
            if (location[piece(index)]<12) top=dir; else bot=dir;
            return new TurningTwist(top, bot);
        } else if (index < 46) {
            return new FlippingTwist(index<43 ? 0 : 1);
        } else {
            throw new IndexOutOfBoundsException("Index must be <46");
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public void twist(Twist twist) {
        if (animating) throw new AnimatingException();
        SquareOneTwist st = (SquareOneTwist)twist;
        if (st instanceof TurningTwist) {
            TurningTwist tt = (TurningTwist)st;
            turn(tt.top, tt.bot);
        } else {
            flip(((FlippingTwist)st).which);
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public void twist(Twist twist, double time) {
        if (animating) throw new AnimatingException();
        SquareOneTwist st = (SquareOneTwist)twist;
        if (time < 0.0) {
            if (st instanceof TurningTwist) {
                TurningTwist tt = (TurningTwist)st;
                time *= Math.max(Math.abs(tt.top), Math.abs(tt.bot)) / 6.0;
            }
            time = Math.abs(time);
        }
        if (time == 0.0) {
            twist(twist);
            return;
        }
        aniTwist = st;
        animating = true;
        aniTime = time;
        aniCurrent = 0;
    }
    
    /**
     * {@inheritDoc}
     */
    public boolean isAnimating() {
        return animating;
    }
    
    /**
     * {@inheritDoc}
     */
    public void advanceAnimation(double time) {
        if (animating) {
            aniCurrent += time;
            if (aniCurrent > aniTime) {
                animating = false;
                twist(aniTwist);
            }
        }
    }
    
    private int piece(int sticker) {
        if (sticker < 16) return sticker/2;
        else return (sticker+8)/3;
    }
    
    //Turns the top and bottom faces clockwise by top*pi/6 and bot*pi/6 resp.
    private void turn(int top, int bot) {
        top = (top%12+12)%12;
        bot = (bot%12+12)%12;
        int[] oldPiece = Arrays.copyOf(piece, 24);
        int i;
        
        for (i=0; i<12; i++) {
            piece[i] = oldPiece[(i+top)%12];
        }
        for (i=0; i<16; i++) {
            if (location[i]<12) {
                location[i] = (location[i]-top+12)%12;
            }
        }
        for (i=0; i<12; i++) {
            piece[(i+bot)%12+12] = oldPiece[i+12];
        }
        for (i=0; i<16; i++) {
            if (location[i]>=12) {
                location[i] = (location[i]+bot)%12+12;
            }
        }
    }
    
    //Tries to flip the ((which & 1) == 0 ? front : back) half of the puzzle.
    //Returns true if it did it.
    private boolean flip(int which) {
        if (piece[0]==piece[11] ||
            piece[6]==piece[5] ||
            piece[12]==piece[23] ||
            piece[18]==piece[17]) return false;
        
        int[] oldPiece = Arrays.copyOf(piece, 24);
        int i;
        
        if ((which & 1) == 0) {
            for (i=0; i<6; i++) {
                piece[i] = oldPiece[17-i];
                piece[17-i] = oldPiece[i];
            }
            for (i=0; i<8; i++) {
                if (location[i]<6 || location[i]>=12 && location[i]<18) location[i]=17-location[i];
            }
            for (i=8; i<16; i++) {
                if (location[i]<6 || location[i]>=12 && location[i]<18) location[i]=16-location[i];
            }
            midFlip[0]^=true;
        } else {
            for (i=0; i<6; i++) {
                piece[6+i] = oldPiece[23-i];
                piece[23-i] = oldPiece[6+i];
            }
            for (i=0; i<8; i++) {
                if (location[i]>=6 && location[i]<12 || location[i]>=18) location[i]=29-location[i];
            }
            for (i=8; i<16; i++) {
                if (location[i]>=6 && location[i]<12 || location[i]>=18) location[i]=28-location[i];
            }
            midFlip[1]^=true;
        }
        return true;
    }
    
    /**
     * {@inheritDoc}
     */
    public void scramble() {
        Random r = new Random();
        int parity = r.nextInt(2);
        for (int i=0; i<20+parity; i++) {
            turn(r.nextInt(12), r.nextInt(12));
            while (piece[0]==piece[11] || piece[6]==piece[5]) turn(r.nextInt(12), 0);
            while (piece[12]==piece[23] || piece[18]==piece[17]) turn(0, r.nextInt(12));
            flip(r.nextInt(2));
        }
    }
    
    /**
     * {@inheritDoc}
     * <br>
     * This puzzle is considered solved if it is in a cube shape and each face
     * is entirely one colour. Equivalently, it is solved if each piece is in
     * its original position, or if each piece is in its original position
     * except that the entire cube has been flipped over.
     * 
     * @return {@inheritDoc}
     */
    public boolean isSolved() {
        if (midFlip[0] != midFlip[1]) return false;
        if (midFlip[0]) {
            for (int i=0; i<16; i++) if (location[i] != 17-i%2*3+i/2%2*6-i/4%2*15-i/8%2*2) return false;
        } else {
            for (int i=0; i<16; i++) if (location[i] != 3*i+i/8) return false;
        }
        return true;
    }
    
}
